"""
Modified from https://github.com/agrimgupta92/sgan/blob/master/sgan/data/trajectories.py.

"""
import logging
import os
import math

import numpy as np

import torch
from torch.utils.data import Dataset

logger = logging.getLogger(__name__)


def seq_collate(data):
    (
        obs_seq_list,
        pred_seq_list,
        obs_seq_rel_list,
        pred_seq_rel_list,
        non_linear_ped_list,
        loss_mask_list,
    ) = zip(*data)

    _len = [len(seq) for seq in obs_seq_list]
    cum_start_idx = [0] + np.cumsum(_len).tolist()
    seq_start_end = [
        [start, end] for start, end in zip(cum_start_idx, cum_start_idx[1:])
    ]

    # Data format: batch, input_size, seq_len
    # LSTM input format: seq_len, batch, input_size
    obs_traj = torch.cat(obs_seq_list, dim=0).permute(2, 0, 1)
    pred_traj = torch.cat(pred_seq_list, dim=0).permute(2, 0, 1)
    obs_traj_rel = torch.cat(obs_seq_rel_list, dim=0).permute(2, 0, 1)
    pred_traj_rel = torch.cat(pred_seq_rel_list, dim=0).permute(2, 0, 1)
    non_linear_ped = torch.cat(non_linear_ped_list)
    loss_mask = torch.cat(loss_mask_list, dim=0)
    seq_start_end = torch.LongTensor(seq_start_end)
    out = [
        obs_traj,
        pred_traj,
        obs_traj_rel,
        pred_traj_rel,
        non_linear_ped,
        loss_mask,
        seq_start_end,
    ]

    return tuple(out)


def read_file(_path, delim="\t"):
    data = []
    if delim == "tab":
        delim = "\t"
    elif delim == "space":
        delim = " "
    with open(_path, "r") as f:
        for line in f:
            line = line.strip().split(delim)
            line = [float(i) for i in line]
            data.append(line)
    return np.asarray(data)


def poly_fit(traj, traj_len, threshold):
    """
    Input:
    - traj: Numpy array of shape (2, traj_len)
    - traj_len: Len of trajectory
    - threshold: Minimum error to be considered for non linear traj
    Output:
    - int: 1 -> Non Linear 0-> Linear
    """
    t = np.linspace(0, traj_len - 1, traj_len)
    res_x = np.polyfit(t, traj[0, -traj_len:], 2, full=True)[1]
    res_y = np.polyfit(t, traj[1, -traj_len:], 2, full=True)[1]
    if res_x + res_y >= threshold:
        return 1.0
    else:
        return 0.0


class TrajectoryDataset(Dataset):
    """Dataloader for the Trajectory datasets"""

    def __init__(
        self,
        data_dir,
        obs_len=8,
        pred_len=12,
        skip=1,
        threshold=0.002,
        min_ped=1,
        delim="\t",
    ):
        """
        Args:
        - data_dir: Directory containing dataset files in the format
        <frame_id> <ped_id> <x> <y>
        - obs_len: Number of time-steps in input trajectories
        - pred_len: Number of time-steps in output trajectories
        - skip: Number of frames to skip while making the dataset
        - threshold: Minimum error to be considered for non linear traj
        when using a linear predictor
        - min_ped: Minimum number of pedestrians that should be in a seqeunce
        - delim: Delimiter in the dataset files
        """
        super(TrajectoryDataset, self).__init__()

        self.data_dir = data_dir
        self.obs_len = obs_len
        self.pred_len = pred_len
        self.skip = skip
        self.seq_len = self.obs_len + self.pred_len
        self.delim = delim

        all_files = os.listdir(self.data_dir)
        all_files = [os.path.join(self.data_dir, _path) for _path in all_files]
        num_peds_in_seq = []
        seq_list = []
        seq_list_rel = []
        loss_mask_list = []
        non_linear_ped = []
        for path in all_files:
            data = read_file(path, delim)
            frames = np.unique(data[:, 0]).tolist()
            frame_data = []
            for frame in frames:
                frame_data.append(data[frame == data[:, 0], :])
            num_sequences = int(math.ceil((len(frames) - self.seq_len + 1) / skip))

            for idx in range(0, num_sequences * self.skip + 1, skip):
                curr_seq_data = np.concatenate(
                    frame_data[idx : idx + self.seq_len], axis=0
                )
                peds_in_curr_seq = np.unique(curr_seq_data[:, 1])
                curr_seq_rel = np.zeros((len(peds_in_curr_seq), 2, self.seq_len))
                curr_seq = np.zeros((len(peds_in_curr_seq), 2, self.seq_len))
                curr_loss_mask = np.zeros((len(peds_in_curr_seq), self.seq_len))
                num_peds_considered = 0
                _non_linear_ped = []
                for _, ped_id in enumerate(peds_in_curr_seq):
                    curr_ped_seq = curr_seq_data[curr_seq_data[:, 1] == ped_id, :]
                    curr_ped_seq = np.around(curr_ped_seq, decimals=4)
                    pad_front = frames.index(curr_ped_seq[0, 0]) - idx
                    pad_end = frames.index(curr_ped_seq[-1, 0]) - idx + 1
                    if pad_end - pad_front != self.seq_len:
                        continue
                    curr_ped_seq = np.transpose(curr_ped_seq[:, 2:])
                    curr_ped_seq = curr_ped_seq
                    # Make coordinates relative
                    rel_curr_ped_seq = np.zeros(curr_ped_seq.shape)
                    rel_curr_ped_seq[:, 1:] = curr_ped_seq[:, 1:] - curr_ped_seq[:, :-1]
                    _idx = num_peds_considered
                    curr_seq[_idx, :, pad_front:pad_end] = curr_ped_seq
                    curr_seq_rel[_idx, :, pad_front:pad_end] = rel_curr_ped_seq
                    # Linear vs Non-Linear Trajectory
                    _non_linear_ped.append(poly_fit(curr_ped_seq, pred_len, threshold))
                    curr_loss_mask[_idx, pad_front:pad_end] = 1
                    num_peds_considered += 1

                if num_peds_considered > min_ped:
                    non_linear_ped += _non_linear_ped
                    num_peds_in_seq.append(num_peds_considered)
                    loss_mask_list.append(curr_loss_mask[:num_peds_considered])
                    seq_list.append(curr_seq[:num_peds_considered])
                    seq_list_rel.append(curr_seq_rel[:num_peds_considered])

        self.num_seq = len(seq_list)
        seq_list = np.concatenate(seq_list, axis=0)
        seq_list_rel = np.concatenate(seq_list_rel, axis=0)
        loss_mask_list = np.concatenate(loss_mask_list, axis=0)
        non_linear_ped = np.asarray(non_linear_ped)

        # Convert numpy -> Torch Tensor
        self.obs_traj = torch.from_numpy(seq_list[:, :, : self.obs_len]).type(
            torch.float
        )
        self.pred_traj = torch.from_numpy(seq_list[:, :, self.obs_len :]).type(
            torch.float
        )
        self.obs_traj_rel = torch.from_numpy(seq_list_rel[:, :, : self.obs_len]).type(
            torch.float
        )
        self.pred_traj_rel = torch.from_numpy(seq_list_rel[:, :, self.obs_len :]).type(
            torch.float
        )
        self.loss_mask = torch.from_numpy(loss_mask_list).type(torch.float)
        self.non_linear_ped = torch.from_numpy(non_linear_ped).type(torch.float)
        cum_start_idx = [0] + np.cumsum(num_peds_in_seq).tolist()
        self.seq_start_end = [
            (start, end) for start, end in zip(cum_start_idx, cum_start_idx[1:])
        ]

    def __len__(self):
        return self.num_seq

    def __getitem__(self, index):
        start, end = self.seq_start_end[index]
        out = [
            self.obs_traj[start:end, :],
            self.pred_traj[start:end, :],
            self.obs_traj_rel[start:end, :],
            self.pred_traj_rel[start:end, :],
            self.non_linear_ped[start:end],
            self.loss_mask[start:end, :],
        ]
        return out



from torch.utils.data import DataLoader

from ..datasets.dataset import TrajectoryDataset, seq_collate


def data_loader(args, path):
    dset = TrajectoryDataset(
        path,
        obs_len=args.obs_len,
        pred_len=args.pred_len,
        skip=args.skip,
        delim=args.delim,
    )

    loader = DataLoader(
        dset,
        batch_size=args.batch_size,
        shuffle=True,
        num_workers=args.loader_num_workers,
        collate_fn=seq_collate,
    )
    return dset, loader



from . import models
from . import datasets

from .accessor import TrajaAccessor
from .frame import TrajaDataFrame, TrajaCollection
from .parsers import read_file, from_df
from .plotting import *
from .trajectory import *

__author__ = "justinshenk"
__version__ = "0.1.7"



import subprocess
import glob
import os
from typing import List

import pandas as pd


def load_ped_datasets() -> List[str]:
    """Returns paths after downloading pedestrian datasets."""
    if not os.path.exists("datasets"):
        subprocess.call(
            ["wget", "https://www.dropbox.com/s/8n02xqv3l9q18r1/datasets.zip"]
        )
        subprocess.call(["unzip", "-q", "datasets.zip"])
        subprocess.call(["rm", "-rf", "datasets.zip"])
    else:
        print("Directory 'datasets' exists, skipping download")

    return glob.glob(f"datasets/*/*")


def load_ped_data(dataset_name=None, aspaths=False) -> dict:
    """Returns pedestrian (ETH, Zara1, Zara2, Univ, Hotel) datasets as dataframe or as paths.

    Args:
        dataset_name: Optional(str) - returns specific dataset
                        eth
                        zara1
                        zara2
                        univ
                        hotel
        aspaths: (bool) - Returns paths only

    Returns:
        paths/dfs (dict) - train/val/test split for paths or dfs, depending on `aspaths` value


    Paths are .txt files with format <frame_id> <ped_id> <x> <y>.
    """
    paths = load_ped_datasets()

    if dataset_name:
        # Get subset of data
        paths = [path for path in paths if dataset_name in path]

    train_dir = [path for path in paths if "train" in path][0]
    val_dir = [path for path in paths if "val" in path][0]
    test_dir = [path for path in paths if "test" in path][0]

    train_paths = glob.glob(os.path.join(train_dir, "*.txt"))
    val_paths = glob.glob(os.path.join(val_dir, "*.txt"))
    test_paths = glob.glob(os.path.join(test_dir, "*.txt"))

    paths = {"train": train_paths, "val": val_paths, "test": test_paths}
    if aspaths:
        return paths

    col_names = ["frame_id", "ped_id", "x", "y"]
    dfs = {
        "train": [pd.read_csv(path, sep="\t", names=col_names) for path in train_paths],
        "val": [pd.read_csv(path, sep="\t", names=col_names) for path in train_paths],
        "test": [pd.read_csv(path, sep="\t", names=col_names) for path in train_paths],
    }
    return dfs



import shapely
import pandas as pd

import traja

df = traja.generate(n=20)


def test_center():
    xy = df.traja.center


def test_night():
    df["time"] = pd.DatetimeIndex(range(20))
    df.traja.night()


def test_between():
    df["time"] = pd.DatetimeIndex(range(20))
    df.traja.between("8:00", "10:00")


def test_day():
    df["time"] = pd.DatetimeIndex(range(20))
    df.traja.day()


def test_xy():
    xy = df.traja.xy
    assert xy.shape == (20, 2)


def test_calc_derivatives():
    df.traja.calc_derivatives()


def test_get_derivatives():
    df.traja.get_derivatives()


def test_speed_intervals():
    si = df.traja.speed_intervals(faster_than=100)
    assert isinstance(si, traja.TrajaDataFrame)


def test_to_shapely():
    shape = df.traja.to_shapely()
    assert isinstance(shape, shapely.geometry.linestring.LineString)


def test_calc_displacement():
    disp = df.traja.calc_displacement()
    assert isinstance(disp, pd.Series)


def test_calc_angle():
    angle = df.traja.calc_angle()
    assert isinstance(angle, pd.Series)


def test_scale():
    df_copy = df.copy()
    df_copy.traja.scale(0.1)
    assert isinstance(df_copy, traja.TrajaDataFrame)


def test_rediscretize(R=0.1):
    df_copy = df.copy()
    r_df = df_copy.traja.rediscretize(R)
    assert isinstance(r_df, traja.TrajaDataFrame)
    assert r_df.shape == (382, 2)


def test_calc_heading():
    heading = df.traja.calc_heading()
    assert isinstance(heading, pd.Series)


def test_calc_turn_angle():
    turn_angle = df.traja.calc_turn_angle()
    assert isinstance(turn_angle, pd.Series)



#! /usr/local/env python3
"""Pytorch visualization code modified from Chad Jensen's implementation
(https://discuss.pytorch.org/t/lstm-for-sequence-prediction/22021/3)."""
import logging

import matplotlib.pyplot as plt
import numpy as np

try:
    import torch
except ImportError:
    raise ImportError(
        "Missing optional dependency 'pytorch'. Install it via pytorch.org"
    )
import torch.nn as nn
import torch.optim as optim


nb_steps = 10

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class LSTM(nn.Module):
    def __init__(self):
        super(LSTM, self).__init__()
        self.lstm = nn.LSTM(2, 100)
        self.head = nn.Linear(100, 2)

    def forward(self, x):
        outputs, states = self.lstm(x)
        outputs = outputs.reshape(x.shape[0] * x.shape[1], -1)
        pred = self.head(outputs)

        return pred


class TrajectoryLSTM:
    def __init__(
        self, xy, nb_steps=10, epochs=1000, batch_size=1, criterion=nn.MSELoss()
    ):
        fig, ax = plt.subplots(2, 1)
        self.fig = fig
        self.ax = ax
        assert xy.shape[1] is 2, f"xy should be an N x 2 array, but is {xy.shape}"
        self.xy = xy
        self.nb_steps = nb_steps
        self.epochs = epochs
        self.batch_size = batch_size
        self.criterion = criterion
        self.rnn = LSTM()

    def load_batch(self, batch_size=32):
        t_1_b = np.zeros((self.nb_steps, self.batch_size, 2))
        t_b = np.zeros((self.nb_steps * self.batch_size, 2))

        inds = np.random.randint(0, len(self.xy) - self.nb_steps, (self.batch_size))
        for i, ind in enumerate(inds):
            t_1_b[:, i] = self.xy[ind : ind + self.nb_steps]
            t_b[i * nb_steps : (i + 1) * self.nb_steps] = self.xy[
                ind + 1 : ind + nb_steps + 1
            ]
        return torch.from_numpy(t_1_b).float(), torch.from_numpy(t_b).float()

    def train(self):
        self.mean_loss = 0.0
        for epoch in range(1, self.epochs + 1):
            t_1_b, t_b = self.load_batch(self.batch_size)

            def closure():
                global loss
                optimizer.zero_grad()
                pred = self.rnn(t_1_b)
                shaped_pred = pred.reshape(-1, 2)
                loss = self.criterion(abs(shaped_pred), abs(t_b))
                loss.backward()

                return loss

            optimizer = optim.Adam(self.rnn.parameters(), 1e-3)
            optimizer.step(closure)
            self.mean_loss += loss.item()

            if epoch % 100 == 0:
                print("Epoch: {} | Loss: {:.6f}".format(epoch, self.mean_loss))
                self.mean_loss = 0

    def savefig(self, filepath):
        self.fig.savefig(filepath)

    def _plot(self):
        t_1_b, t_b = self.load_batch(1)
        pred = self.rnn(t_1_b).detach().numpy().reshape(-1, 2)

        real = t_1_b.numpy().reshape(-1, 2)
        x, y = self.xy.T
        self.ax[0].plot(x, y, label="Real")
        self.ax[0].plot(real[:, 0], real[:, 1], label="Real batch")
        self.ax[0].plot(pred[:, 0], pred[:, 1], label="Pred")

        self.ax[1].scatter(real[:, 0], real[:, 1], label="Real")
        self.ax[1].scatter(pred[:, 0], pred[:, 1], label="Pred")

        for a in self.ax:
            a.legend()

    def plot(self, interactive=True):
        if interactive and (plt.get_backend() == "agg"):
            logging.ERROR("Not able to use interactive plotting in mpl `agg` mode.")
            # interactive = False
        elif interactive:
            while True:
                for a in self.ax:
                    a.clear()
                self._plot()
                plt.pause(1)
                plt.show(block=False)
        else:
            self._plot()
            return self.fig


import torch
import torch.nn as nn


def make_mlp(dim_list, activation="relu", batch_norm=True, dropout=0):
    layers = []
    for dim_in, dim_out in zip(dim_list[:-1], dim_list[1:]):
        layers.append(nn.Linear(dim_in, dim_out))
        if batch_norm:
            layers.append(nn.BatchNorm1d(dim_out))
        if activation == "relu":
            layers.append(nn.ReLU())
        elif activation == "leakyrelu":
            layers.append(nn.LeakyReLU())
        if dropout > 0:
            layers.append(nn.Dropout(p=dropout))
    return nn.Sequential(*layers)


def get_noise(shape, noise_type):
    if noise_type == "gaussian":
        return torch.randn(*shape).cuda()
    elif noise_type == "uniform":
        return torch.rand(*shape).sub_(0.5).mul_(2.0).cuda()
    raise ValueError('Unrecognized noise type "%s"' % noise_type)


class Encoder(nn.Module):
    """Encoder is part of both TrajectoryGenerator and
    TrajectoryDiscriminator"""

    def __init__(
        self, embedding_dim=64, h_dim=64, mlp_dim=1024, num_layers=1, dropout=0.0
    ):
        super(Encoder, self).__init__()

        self.mlp_dim = 1024
        self.h_dim = h_dim
        self.embedding_dim = embedding_dim
        self.num_layers = num_layers

        self.encoder = nn.LSTM(embedding_dim, h_dim, num_layers, dropout=dropout)

        self.spatial_embedding = nn.Linear(2, embedding_dim)

    def init_hidden(self, batch):
        return (
            torch.zeros(self.num_layers, batch, self.h_dim).cuda(),
            torch.zeros(self.num_layers, batch, self.h_dim).cuda(),
        )

    def forward(self, obs_traj):
        """
        Inputs:
        - obs_traj: Tensor of shape (obs_len, batch, 2)
        Output:
        - final_h: Tensor of shape (self.num_layers, batch, self.h_dim)
        """
        # Encode observed Trajectory
        batch = obs_traj.size(1)
        obs_traj_embedding = self.spatial_embedding(obs_traj.view(-1, 2))
        obs_traj_embedding = obs_traj_embedding.view(-1, batch, self.embedding_dim)
        state_tuple = self.init_hidden(batch)
        output, state = self.encoder(obs_traj_embedding, state_tuple)
        final_h = state[0]
        return final_h


class Decoder(nn.Module):
    """Decoder is part of TrajectoryGenerator"""

    def __init__(
        self,
        seq_len,
        embedding_dim=64,
        h_dim=128,
        mlp_dim=1024,
        num_layers=1,
        pool_every_timestep=True,
        dropout=0.0,
        bottleneck_dim=1024,
        activation="relu",
        batch_norm=True,
        pooling_type="pool_net",
        neighborhood_size=2.0,
        grid_size=8,
    ):
        super(Decoder, self).__init__()

        self.seq_len = seq_len
        self.mlp_dim = mlp_dim
        self.h_dim = h_dim
        self.embedding_dim = embedding_dim
        self.pool_every_timestep = pool_every_timestep

        self.decoder = nn.LSTM(embedding_dim, h_dim, num_layers, dropout=dropout)

        if pool_every_timestep:
            if pooling_type == "pool_net":
                self.pool_net = PoolHiddenNet(
                    embedding_dim=self.embedding_dim,
                    h_dim=self.h_dim,
                    mlp_dim=mlp_dim,
                    bottleneck_dim=bottleneck_dim,
                    activation=activation,
                    batch_norm=batch_norm,
                    dropout=dropout,
                )
            elif pooling_type == "spool":
                self.pool_net = SocialPooling(
                    h_dim=self.h_dim,
                    activation=activation,
                    batch_norm=batch_norm,
                    dropout=dropout,
                    neighborhood_size=neighborhood_size,
                    grid_size=grid_size,
                )

            mlp_dims = [h_dim + bottleneck_dim, mlp_dim, h_dim]
            self.mlp = make_mlp(
                mlp_dims, activation=activation, batch_norm=batch_norm, dropout=dropout
            )

        self.spatial_embedding = nn.Linear(2, embedding_dim)
        self.hidden2pos = nn.Linear(h_dim, 2)

    def forward(self, last_pos, last_pos_rel, state_tuple, seq_start_end):
        """
        Inputs:
        - last_pos: Tensor of shape (batch, 2)
        - last_pos_rel: Tensor of shape (batch, 2)
        - state_tuple: (hh, ch) each tensor of shape (num_layers, batch, h_dim)
        - seq_start_end: A list of tuples which delimit sequences within batch
        Output:
        - pred_traj: tensor of shape (self.seq_len, batch, 2)
        """
        batch = last_pos.size(0)
        pred_traj_fake_rel = []
        decoder_input = self.spatial_embedding(last_pos_rel)
        decoder_input = decoder_input.view(1, batch, self.embedding_dim)

        for _ in range(self.seq_len):
            output, state_tuple = self.decoder(decoder_input, state_tuple)
            rel_pos = self.hidden2pos(output.view(-1, self.h_dim))
            curr_pos = rel_pos + last_pos

            if self.pool_every_timestep:
                decoder_h = state_tuple[0]
                pool_h = self.pool_net(decoder_h, seq_start_end, curr_pos)
                decoder_h = torch.cat([decoder_h.view(-1, self.h_dim), pool_h], dim=1)
                decoder_h = self.mlp(decoder_h)
                decoder_h = torch.unsqueeze(decoder_h, 0)
                state_tuple = (decoder_h, state_tuple[1])

            embedding_input = rel_pos

            decoder_input = self.spatial_embedding(embedding_input)
            decoder_input = decoder_input.view(1, batch, self.embedding_dim)
            pred_traj_fake_rel.append(rel_pos.view(batch, -1))
            last_pos = curr_pos

        pred_traj_fake_rel = torch.stack(pred_traj_fake_rel, dim=0)
        return pred_traj_fake_rel, state_tuple[0]


class PoolHiddenNet(nn.Module):
    """Pooling module as proposed in our paper"""

    def __init__(
        self,
        embedding_dim=64,
        h_dim=64,
        mlp_dim=1024,
        bottleneck_dim=1024,
        activation="relu",
        batch_norm=True,
        dropout=0.0,
    ):
        super(PoolHiddenNet, self).__init__()

        self.mlp_dim = 1024
        self.h_dim = h_dim
        self.bottleneck_dim = bottleneck_dim
        self.embedding_dim = embedding_dim

        mlp_pre_dim = embedding_dim + h_dim
        mlp_pre_pool_dims = [mlp_pre_dim, 512, bottleneck_dim]

        self.spatial_embedding = nn.Linear(2, embedding_dim)
        self.mlp_pre_pool = make_mlp(
            mlp_pre_pool_dims,
            activation=activation,
            batch_norm=batch_norm,
            dropout=dropout,
        )

    def repeat(self, tensor, num_reps):
        """
        Inputs:
        -tensor: 2D tensor of any shape
        -num_reps: Number of times to repeat each row
        Outpus:
        -repeat_tensor: Repeat each row such that: R1, R1, R2, R2
        """
        col_len = tensor.size(1)
        tensor = tensor.unsqueeze(dim=1).repeat(1, num_reps, 1)
        tensor = tensor.view(-1, col_len)
        return tensor

    def forward(self, h_states, seq_start_end, end_pos):
        """
        Inputs:
        - h_states: Tensor of shape (num_layers, batch, h_dim)
        - seq_start_end: A list of tuples which delimit sequences within batch
        - end_pos: Tensor of shape (batch, 2)
        Output:
        - pool_h: Tensor of shape (batch, bottleneck_dim)
        """
        pool_h = []
        for _, (start, end) in enumerate(seq_start_end):
            start = start.item()
            end = end.item()
            num_ped = end - start
            curr_hidden = h_states.view(-1, self.h_dim)[start:end]
            curr_end_pos = end_pos[start:end]
            # Repeat -> H1, H2, H1, H2
            curr_hidden_1 = curr_hidden.repeat(num_ped, 1)
            # Repeat position -> P1, P2, P1, P2
            curr_end_pos_1 = curr_end_pos.repeat(num_ped, 1)
            # Repeat position -> P1, P1, P2, P2
            curr_end_pos_2 = self.repeat(curr_end_pos, num_ped)
            curr_rel_pos = curr_end_pos_1 - curr_end_pos_2
            curr_rel_embedding = self.spatial_embedding(curr_rel_pos)
            mlp_h_input = torch.cat([curr_rel_embedding, curr_hidden_1], dim=1)
            curr_pool_h = self.mlp_pre_pool(mlp_h_input)
            curr_pool_h = curr_pool_h.view(num_ped, num_ped, -1).max(1)[0]
            pool_h.append(curr_pool_h)
        pool_h = torch.cat(pool_h, dim=0)
        return pool_h


class SocialPooling(nn.Module):
    """Current state of the art pooling mechanism:
    http://cvgl.stanford.edu/papers/CVPR16_Social_LSTM.pdf"""

    def __init__(
        self,
        h_dim=64,
        activation="relu",
        batch_norm=True,
        dropout=0.0,
        neighborhood_size=2.0,
        grid_size=8,
        pool_dim=None,
    ):
        super(SocialPooling, self).__init__()
        self.h_dim = h_dim
        self.grid_size = grid_size
        self.neighborhood_size = neighborhood_size
        if pool_dim:
            mlp_pool_dims = [grid_size * grid_size * h_dim, pool_dim]
        else:
            mlp_pool_dims = [grid_size * grid_size * h_dim, h_dim]

        self.mlp_pool = make_mlp(
            mlp_pool_dims, activation=activation, batch_norm=batch_norm, dropout=dropout
        )

    def get_bounds(self, ped_pos):
        top_left_x = ped_pos[:, 0] - self.neighborhood_size / 2
        top_left_y = ped_pos[:, 1] + self.neighborhood_size / 2
        bottom_right_x = ped_pos[:, 0] + self.neighborhood_size / 2
        bottom_right_y = ped_pos[:, 1] - self.neighborhood_size / 2
        top_left = torch.stack([top_left_x, top_left_y], dim=1)
        bottom_right = torch.stack([bottom_right_x, bottom_right_y], dim=1)
        return top_left, bottom_right

    def get_grid_locations(self, top_left, other_pos):
        cell_x = torch.floor(
            ((other_pos[:, 0] - top_left[:, 0]) / self.neighborhood_size)
            * self.grid_size
        )
        cell_y = torch.floor(
            ((top_left[:, 1] - other_pos[:, 1]) / self.neighborhood_size)
            * self.grid_size
        )
        grid_pos = cell_x + cell_y * self.grid_size
        return grid_pos

    def repeat(self, tensor, num_reps):
        """
        Inputs:
        -tensor: 2D tensor of any shape
        -num_reps: Number of times to repeat each row
        Outpus:
        -repeat_tensor: Repeat each row such that: R1, R1, R2, R2
        """
        col_len = tensor.size(1)
        tensor = tensor.unsqueeze(dim=1).repeat(1, num_reps, 1)
        tensor = tensor.view(-1, col_len)
        return tensor

    def forward(self, h_states, seq_start_end, end_pos):
        """
        Inputs:
        - h_states: Tesnsor of shape (num_layers, batch, h_dim)
        - seq_start_end: A list of tuples which delimit sequences within batch.
        - end_pos: Absolute end position of obs_traj (batch, 2)
        Output:
        - pool_h: Tensor of shape (batch, h_dim)
        """
        pool_h = []
        for _, (start, end) in enumerate(seq_start_end):
            start = start.item()
            end = end.item()
            num_ped = end - start
            grid_size = self.grid_size * self.grid_size
            curr_hidden = h_states.view(-1, self.h_dim)[start:end]
            curr_hidden_repeat = curr_hidden.repeat(num_ped, 1)
            curr_end_pos = end_pos[start:end]
            curr_pool_h_size = (num_ped * grid_size) + 1
            curr_pool_h = curr_hidden.new_zeros((curr_pool_h_size, self.h_dim))
            # curr_end_pos = curr_end_pos.data
            top_left, bottom_right = self.get_bounds(curr_end_pos)

            # Repeat position -> P1, P2, P1, P2
            curr_end_pos = curr_end_pos.repeat(num_ped, 1)
            # Repeat bounds -> B1, B1, B2, B2
            top_left = self.repeat(top_left, num_ped)
            bottom_right = self.repeat(bottom_right, num_ped)

            grid_pos = self.get_grid_locations(top_left, curr_end_pos).type_as(
                seq_start_end
            )
            # Make all positions to exclude as non-zero
            # Find which peds to exclude
            x_bound = (curr_end_pos[:, 0] >= bottom_right[:, 0]) + (
                curr_end_pos[:, 0] <= top_left[:, 0]
            )
            y_bound = (curr_end_pos[:, 1] >= top_left[:, 1]) + (
                curr_end_pos[:, 1] <= bottom_right[:, 1]
            )

            within_bound = x_bound + y_bound
            within_bound[0 :: num_ped + 1] = 1  # Don't include the ped itself
            within_bound = within_bound.view(-1)

            # This is a tricky way to get scatter add to work. Helps me avoid a
            # for loop. Offset everything by 1. Use the initial 0 position to
            # dump all uncessary adds.
            grid_pos += 1
            total_grid_size = self.grid_size * self.grid_size
            offset = torch.arange(
                0, total_grid_size * num_ped, total_grid_size
            ).type_as(seq_start_end)

            offset = self.repeat(offset.view(-1, 1), num_ped).view(-1)
            grid_pos += offset
            grid_pos[within_bound != 0] = 0
            grid_pos = grid_pos.view(-1, 1).expand_as(curr_hidden_repeat)

            curr_pool_h = curr_pool_h.scatter_add(0, grid_pos, curr_hidden_repeat)
            curr_pool_h = curr_pool_h[1:]
            pool_h.append(curr_pool_h.view(num_ped, -1))

        pool_h = torch.cat(pool_h, dim=0)
        pool_h = self.mlp_pool(pool_h)
        return pool_h


class TrajectoryGenerator(nn.Module):
    """Modified from @agrimgupta92's https://github.com/agrimgupta92/sgan/blob/master/sgan/models.py."""

    def __init__(
        self,
        obs_len,
        pred_len,
        embedding_dim=64,
        encoder_h_dim=64,
        decoder_h_dim=128,
        mlp_dim=1024,
        num_layers=1,
        noise_dim=(0,),
        noise_type="gaussian",
        noise_mix_type="ped",
        pooling_type=None,
        pool_every_timestep=True,
        dropout=0.0,
        bottleneck_dim=1024,
        activation="relu",
        batch_norm=True,
        neighborhood_size=2.0,
        grid_size=8,
    ):
        super(TrajectoryGenerator, self).__init__()

        if pooling_type and pooling_type.lower() == "none":
            pooling_type = None

        self.obs_len = obs_len
        self.pred_len = pred_len
        self.mlp_dim = mlp_dim
        self.encoder_h_dim = encoder_h_dim
        self.decoder_h_dim = decoder_h_dim
        self.embedding_dim = embedding_dim
        self.noise_dim = noise_dim
        self.num_layers = num_layers
        self.noise_type = noise_type
        self.noise_mix_type = noise_mix_type
        self.pooling_type = pooling_type
        self.noise_first_dim = 0
        self.pool_every_timestep = pool_every_timestep
        self.bottleneck_dim = 1024

        self.encoder = Encoder(
            embedding_dim=embedding_dim,
            h_dim=encoder_h_dim,
            mlp_dim=mlp_dim,
            num_layers=num_layers,
            dropout=dropout,
        )

        self.decoder = Decoder(
            pred_len,
            embedding_dim=embedding_dim,
            h_dim=decoder_h_dim,
            mlp_dim=mlp_dim,
            num_layers=num_layers,
            pool_every_timestep=pool_every_timestep,
            dropout=dropout,
            bottleneck_dim=bottleneck_dim,
            activation=activation,
            batch_norm=batch_norm,
            pooling_type=pooling_type,
            grid_size=grid_size,
            neighborhood_size=neighborhood_size,
        )

        if pooling_type == "pool_net":
            self.pool_net = PoolHiddenNet(
                embedding_dim=self.embedding_dim,
                h_dim=encoder_h_dim,
                mlp_dim=mlp_dim,
                bottleneck_dim=bottleneck_dim,
                activation=activation,
                batch_norm=batch_norm,
            )
        elif pooling_type == "spool":
            self.pool_net = SocialPooling(
                h_dim=encoder_h_dim,
                activation=activation,
                batch_norm=batch_norm,
                dropout=dropout,
                neighborhood_size=neighborhood_size,
                grid_size=grid_size,
            )

        if self.noise_dim[0] == 0:
            self.noise_dim = None
        else:
            self.noise_first_dim = noise_dim[0]

        # Decoder Hidden
        if pooling_type:
            input_dim = encoder_h_dim + bottleneck_dim
        else:
            input_dim = encoder_h_dim

        if self.mlp_decoder_needed():
            mlp_decoder_context_dims = [
                input_dim,
                mlp_dim,
                decoder_h_dim - self.noise_first_dim,
            ]

            self.mlp_decoder_context = make_mlp(
                mlp_decoder_context_dims,
                activation=activation,
                batch_norm=batch_norm,
                dropout=dropout,
            )

    def add_noise(self, _input, seq_start_end, user_noise=None):
        """
        Inputs:
        - _input: Tensor of shape (_, decoder_h_dim - noise_first_dim)
        - seq_start_end: A list of tuples which delimit sequences within batch.
        - user_noise: Generally used for inference when you want to see
        relation between different types of noise and outputs.
        Outputs:
        - decoder_h: Tensor of shape (_, decoder_h_dim)
        """
        if not self.noise_dim:
            return _input

        if self.noise_mix_type == "global":
            noise_shape = (seq_start_end.size(0),) + self.noise_dim
        else:
            noise_shape = (_input.size(0),) + self.noise_dim

        if user_noise is not None:
            z_decoder = user_noise
        else:
            z_decoder = get_noise(noise_shape, self.noise_type)

        if self.noise_mix_type == "global":
            _list = []
            for idx, (start, end) in enumerate(seq_start_end):
                start = start.item()
                end = end.item()
                _vec = z_decoder[idx].view(1, -1)
                _to_cat = _vec.repeat(end - start, 1)
                _list.append(torch.cat([_input[start:end], _to_cat], dim=1))
            decoder_h = torch.cat(_list, dim=0)
            return decoder_h

        decoder_h = torch.cat([_input, z_decoder], dim=1)

        return decoder_h

    def mlp_decoder_needed(self):
        if (
            self.noise_dim
            or self.pooling_type
            or self.encoder_h_dim != self.decoder_h_dim
        ):
            return True
        else:
            return False

    def forward(self, obs_traj, obs_traj_rel, seq_start_end, user_noise=None):
        """
        Inputs:
        - obs_traj: Tensor of shape (obs_len, batch, 2)
        - obs_traj_rel: Tensor of shape (obs_len, batch, 2)
        - seq_start_end: A list of tuples which delimit sequences within batch.
        - user_noise: Generally used for inference when you want to see
        relation between different types of noise and outputs.
        Output:
        - pred_traj_rel: Tensor of shape (self.pred_len, batch, 2)
        """
        batch = obs_traj_rel.size(1)
        # Encode seq
        final_encoder_h = self.encoder(obs_traj_rel)
        # Pool States
        if self.pooling_type:
            end_pos = obs_traj[-1, :, :]
            pool_h = self.pool_net(final_encoder_h, seq_start_end, end_pos)
            # Construct input hidden states for decoder
            mlp_decoder_context_input = torch.cat(
                [final_encoder_h.view(-1, self.encoder_h_dim), pool_h], dim=1
            )
        else:
            mlp_decoder_context_input = final_encoder_h.view(-1, self.encoder_h_dim)

        # Add Noise
        if self.mlp_decoder_needed():
            noise_input = self.mlp_decoder_context(mlp_decoder_context_input)
        else:
            noise_input = mlp_decoder_context_input
        decoder_h = self.add_noise(noise_input, seq_start_end, user_noise=user_noise)
        decoder_h = torch.unsqueeze(decoder_h, 0)

        decoder_c = torch.zeros(self.num_layers, batch, self.decoder_h_dim).cuda()

        state_tuple = (decoder_h, decoder_c)
        last_pos = obs_traj[-1]
        last_pos_rel = obs_traj_rel[-1]
        # Predict Trajectory

        decoder_out = self.decoder(last_pos, last_pos_rel, state_tuple, seq_start_end)
        pred_traj_fake_rel, final_decoder_h = decoder_out

        return pred_traj_fake_rel


class TrajectoryDiscriminator(nn.Module):
    def __init__(
        self,
        obs_len,
        pred_len,
        embedding_dim=64,
        h_dim=64,
        mlp_dim=1024,
        num_layers=1,
        activation="relu",
        batch_norm=True,
        dropout=0.0,
        d_type="local",
    ):
        super(TrajectoryDiscriminator, self).__init__()

        self.obs_len = obs_len
        self.pred_len = pred_len
        self.seq_len = obs_len + pred_len
        self.mlp_dim = mlp_dim
        self.h_dim = h_dim
        self.d_type = d_type

        self.encoder = Encoder(
            embedding_dim=embedding_dim,
            h_dim=h_dim,
            mlp_dim=mlp_dim,
            num_layers=num_layers,
            dropout=dropout,
        )

        real_classifier_dims = [h_dim, mlp_dim, 1]
        self.real_classifier = make_mlp(
            real_classifier_dims,
            activation=activation,
            batch_norm=batch_norm,
            dropout=dropout,
        )
        if d_type == "global":
            mlp_pool_dims = [h_dim + embedding_dim, mlp_dim, h_dim]
            self.pool_net = PoolHiddenNet(
                embedding_dim=embedding_dim,
                h_dim=h_dim,
                mlp_dim=mlp_pool_dims,
                bottleneck_dim=h_dim,
                activation=activation,
                batch_norm=batch_norm,
            )

    def forward(self, traj, traj_rel, seq_start_end=None):
        """
        Inputs:
        - traj: Tensor of shape (obs_len + pred_len, batch, 2)
        - traj_rel: Tensor of shape (obs_len + pred_len, batch, 2)
        - seq_start_end: A list of tuples which delimit sequences within batch
        Output:
        - scores: Tensor of shape (batch,) with real/fake scores
        """
        final_h = self.encoder(traj_rel)
        # Note: In case of 'global' option we are using start_pos as opposed to
        # end_pos. The intution being that hidden state has the whole
        # trajectory and relative postion at the start when combined with
        # trajectory information should help in discriminative behavior.
        if self.d_type == "local":
            classifier_input = final_h.squeeze()
        else:
            classifier_input = self.pool_net(final_h.squeeze(), seq_start_end, traj[0])
        scores = self.real_classifier(classifier_input)
        return scores



try:
    import rpy2
except ImportError:
    raise ImportError(
        "Python package rpy2 is required for this. Install it with"
        "\n"
        "pip install rpy2"
        ""
    )
import rpy2.robjects.packages as rpackages

try:
    import rpy2.robjects.pandas2ri as rpandas
except ModuleNotFoundError as e:
    if "tzlocal" in e.msg:
        raise ModuleNotFoundError(
            e.msg + "\n Install tzlocal with `pip install tzlocal`."
        )
    else:
        raise ModuleNotFoundError(e)
from rpy2.robjects.packages import importr


__all__ = ["import_adehabitat", "import_trajr", "plot_ltraj", "to_trajr", "to_ltraj"]


rpandas.activate()

ADEHABITAT_INSTALLED = False
TRAJR_INSTALLED = False


def import_adehabitat(suppress_messages=True):
    global ADEHABITAT_INSTALLED
    if not ADEHABITAT_INSTALLED:
        utils = rpackages.importr("utils", suppress_messages=suppress_messages)
        print("Importing adehabitat")
        utils.chooseCRANmirror(ind=1)
        utils.install_packages("adehabitatLT")
        ADEHABITAT_INSTALLED = True
    adehabitat = importr("adehabitatLT", suppress_messages=suppress_messages)
    return adehabitat


def import_trajr(suppress_messages=True):
    global TRAJR_INSTALLED
    if not TRAJR_INSTALLED:
        utils = rpackages.importr("utils", suppress_messages=suppress_messages)
        print("Importing trajr")
        utils.chooseCRANmirror(ind=1)
        utils.install_packages("trajr")
        TRAJR_INSTALLED = True
    trajr = importr("trajr")
    return trajr


def plot_ltraj(ltraj, id=1):
    """Plot `ltraj` using R method."""
    adehabitat = import_adehabitat()
    adehabitat.plot_ltraj(ltraj, id=1)


def to_trajr(trj):
    """Convert trajectory to R `trajr` object. Default fps is 30.

    Args:
        trajectory (:class:`~traja.main.TrajaDataFrame`): trajectory


    Returns:
        traj (:class:`rpy2.robjects.vectors.DataFrame`): column names are ['x', 'y', 'time', 'displacementTime',
                                                            'polar', 'displacement']

    .. doctest::

        >>> import traja
        >>> df = traja.TrajaDataFrame({'x':range(5),'y':range(5)})
        >>> trjr_df = traja.rutils.to_trajr(df) # doctest: +SKIP
        >>> [x for x in trjr_df.names] # doctest: +SKIP
        ...
        ['x', 'y', 'id', 'time', 'displacementTime', 'polar', 'displacement']


    """
    from traja.trajectory import _get_time_col

    trajr = import_trajr()
    if "id" not in trj.__dict__.keys():
        trj["id"] = 0
    time_col = _get_time_col(trj)
    if time_col == "index":
        trj["time"] = trj.index
        time_col = "time"
    fps = trj.fps
    spatial_units = trj.spatial_units or "m"
    time_units = trj.time_units or "s"

    trj_rdf = rpandas.py2rpy(trj)

    trajr_trj = trajr.TrajFromCoords(
        trj_rdf,
        xCol="x",
        yCol="y",
        timeCol=time_col or rpy2.rinterface.NULL,
        fps=fps or 30,
        spatialUnits=spatial_units,
        timeUnits=time_units,
    )

    return trajr_trj


def to_ltraj(trj, id=1, typeII=False):
    """Convert trajectory to R `ltraj` object.

    Args:
        trajectory (:class:`~traja.main.TrajaDataFrame`): trajectory
        id (str, optional): id of animal/target
        typeII (bool):  contains datatime format time series (not yet supported)

    Returns:
        ltraj (:class:`rpy2.robjects.vectors.ListVector`): first index (ie, ltraj[0]) is pandas DataFrame including
                                                            columns 'x', 'y', 'date', 'dx', 'dy', 'dist', 'dt',
                                                            'R2n', 'abs.angle', 'rel.angle'

    .. doctest::

        >>> import traja
        >>> df = traja.TrajaDataFrame({'x':range(5),'y':range(5)})
        >>> ltraj = traja.rutils.to_ltraj(df) # doctest: +SKIP
        >>> print(ltraj[0]) # doctest: +SKIP
        ...
           x  y  date          dx    ...              dt   R2n  abs.angle  rel.angle
        0  0  0     1           1    ...               1   0.0   0.785398        NaN
        1  1  1     2           1    ...               1   2.0   0.785398        0.0
        2  2  2     3           1    ...               1   8.0   0.785398        0.0
        3  3  3     4           1    ...               1  18.0   0.785398        0.0
        4  4  4     5 -2147483648    ...     -2147483648  32.0        NaN        NaN

        [5 rows x 10 columns]

    """
    adehabitat = import_adehabitat()

    df = trj[["x", "y"]]

    if typeII:
        raise NotImplementedError(
            "datetime series not yet implemented for this method."
        )
        # FIXME: Add date converted from rpy2.robjects.POSIXct
        # date = None
        # ltraj = adehabitat.as_ltraj(df, id=id, date=date, typeII=True)  # Doesn't work
    else:
        ltraj = adehabitat.as_ltraj(df, id=id, typeII=False)[0]
    return ltraj



import os
import shutil
import tempfile

import pandas as pd
from pandas import DataFrame

import traja
from traja import TrajaDataFrame, read_file, TrajaCollection


class TestDataFrame:
    def setup_method(self):
        dirname = os.path.dirname(traja.__file__)
        data_filename = os.path.join(dirname, "tests/data/3527.csv")
        df = read_file(data_filename)
        self.df = read_file(data_filename, xlim=(df.x.min(), df.x.max()))
        self.tempdir = tempfile.mkdtemp()

    def teardown_method(self):
        shutil.rmtree(self.tempdir)

    def test_df_init(self):
        assert isinstance(self.df, TrajaDataFrame)

    # def test_copy(self):
    #     df2 = self.df.copy()
    #     assert (df2, TrajaDataFrame)
    #     assert df2.xlim == self.df.xlim

    def test_dataframe_to_trajadataframe(self):
        df = DataFrame(
            {"x": range(len(self.df)), "y": range(len(self.df))}, index=self.df.index
        )

        tf = TrajaDataFrame(df)
        assert isinstance(df, DataFrame)
        assert isinstance(tf, TrajaDataFrame)

    def test_construct_dataframe(self):
        df = traja.TrajaDataFrame(
            {"x": range(len(self.df)), "y": range(len(self.df))},
            index=self.df.index,
            xlim=(0, 2),
            ylim=(0, 2),
            spatial_units="m",
            title="Serious title",
            fps=2.0,
            time_units="s",
            id=42,
        )

        assert df.title == "Serious title"

        # Test 'merge'
        df2 = df.copy()
        assert df2.title == "Serious title"

        assert df._get_time_col() == None
        assert self.df._get_time_col() == "Time"

        # Modify metavar
        df.set("title", "New title")
        assert df.title == "New title"

        # Test __finalize__
        df_copy = df.copy()
        df2_copy = df2.copy()
        assert isinstance(df_copy, traja.TrajaDataFrame)


class TestTrajaCollection:
    def setup_method(self):
        dirname = os.path.dirname(traja.__file__)
        data_filename = os.path.join(dirname, "tests/data/3527.csv")
        df = read_file(data_filename)
        df = read_file(data_filename, xlim=(df.x.min(), df.x.max()))
        df2 = df.copy()
        df2["TrackId"] = 2
        df2["x"] += 10
        self.coll = TrajaCollection({1: df, 2: df2}, id_col="TrackId")
        self.tempdir = tempfile.mkdtemp()

    def teardown_method(self):
        shutil.rmtree(self.tempdir)

    def test_collection_init(self):
        assert isinstance(self.coll, TrajaCollection)

    # def test_copy(self):
    #     trjs = self.trjs.copy()
    #     assert (trjs, TrajaCollection)
    #     assert trjs.xlim == self.trjs.xlim

    def test_plot_collection(self):
        self.coll.plot()
        # Test with colors
        self.coll.plot(colors={1: "red", 2: "blue"})

    def test_apply_all(self):
        angles = self.coll.apply_all(traja.calc_angle)
        assert isinstance(angles, pd.DataFrame)

        # Test with multiple ids
        coll_copy = self.coll.copy()
        coll_copy.loc[coll_copy.index[:5], "TrackId"] = 1
        angles = coll_copy.apply_all(traja.calc_angle)
        assert isinstance(angles, pd.Series)



from collections import OrderedDict
from datetime import timedelta
from typing import Union, Optional, Tuple, List

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.collections import PathCollection
from matplotlib import dates as md
from matplotlib.figure import Figure
import numpy as np
import pandas as pd
from pandas.core.dtypes.common import (
    is_datetime_or_timedelta_dtype,
    is_datetime64_any_dtype,
    is_timedelta64_dtype,
)

import traja
from traja.frame import TrajaDataFrame
from traja.trajectory import coords_to_flow


__all__ = [
    "_get_after_plot_args",
    "_label_axes",
    "_polar_bar",
    "_process_after_plot_args",
    "animate",
    "bar_plot",
    "color_dark",
    "fill_ci",
    "find_runs",
    "plot",
    "plot_3d",
    "plot_actogram",
    "plot_collection",
    "plot_contour",
    "plot_clustermap",
    "plot_flow",
    "plot_quiver",
    "plot_stream",
    "plot_surface",
    "plot_transition_graph",
    "plot_transition_matrix",
    "plot_xy",
    "polar_bar",
    "predict",
    "sans_serif",
    "stylize_axes",
    "trip_grid",
]


def stylize_axes(ax):
    """Add top and right border to plot, set ticks."""
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)

    ax.xaxis.set_tick_params(top="off", direction="out", width=1)
    ax.yaxis.set_tick_params(right="off", direction="out", width=1)


def sans_serif():
    """Convenience function for changing plot text to serif font."""
    plt.rc("font", family="serif")


def _rolling(df, window, step):
    count = 0
    df_length = len(df)
    while count < (df_length - window):
        yield count, df[count : window + count]
        count += step


def predict(
    xy: np.ndarray,
    nb_steps: int = 10,
    epochs: int = 1000,
    batch_size: int = 1,
    model="lstm",
):  # pragma: no cover
    """Method for training and visualizing LSTM with trajectory datasets."""
    if model is "lstm":
        from traja.models.nn import TrajectoryLSTM

        TrajectoryLSTM(xy, nb_steps=nb_steps, epochs=epochs, batch_size=batch_size)


def bar_plot(trj: TrajaDataFrame, bins: Union[int, tuple] = None, **kwargs) -> Axes:
    """Plot trajectory for single animal over period.

    Args:
      trj (:class:`traja.TrajaDataFrame`): trajectory
      bins (int or tuple): number of bins for x and y
      **kwargs: additional keyword arguments to :meth:`mpl_toolkits.mplot3d.Axed3D.plot`

    Returns:
        ax (:class:`~matplotlib.collections.PathCollection`): Axes of plot

    """
    # TODO: Add time component
    from mpl_toolkits.mplot3d import Axes3D

    bins = traja.trajectory._bins_to_tuple(trj, bins)

    X, Y, U, V = coords_to_flow(trj, bins)

    hist, _ = trip_grid(trj, bins, hist_only=True)
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.set_aspect("equal")
    X = X.flatten("F")
    Y = Y.flatten("F")
    ax.bar3d(
        X,
        Y,
        np.zeros_like(X),
        1,
        1,
        hist.flatten(),
        zsort="average",
        shade=True,
        **kwargs,
    )
    ax.set(xlabel="x", ylabel="y", zlabel="Frames")

    return ax


def plot_rolling_hull(trj: TrajaDataFrame, window=100, step=20, areas=False, **kwargs):
    """Plot rolling convex hull of trajectory. If `areas` is True, only
    areas over time is plotted.
    
    """
    hulls = []

    for offset, window in _rolling(trj, window=window, step=step):
        if window.dropna().empty:
            continue
        shape = window.traja.to_shapely()
        hull = shape.convex_hull
        hulls.append(hull)

    if areas:
        hull_areas = []
        for idx, hull in enumerate(hulls):
            hull_areas.append(hull.area)
        plt.plot(hull_areas, **kwargs)
        plt.title(f"Rolling Trajectory Convex Hull Area\nWindow={window},Step={step}")
        plt.ylabel(f"Area {trj.__dict__.get('spatial_units','m')}")
        plt.xlabel("Frame")
    else:
        xlim, ylim = traja.trajectory._get_xylim(trj)
        plt.xlim = xlim
        plt.ylim = ylim
        for idx, hull in enumerate(hulls):
            if hasattr(
                hull, "exterior"
            ):  # Occassionally a Point object without it reaches
                plt.plot(*hull.exterior.xy, alpha=idx / len(hulls), c="k", **kwargs)
        ax = plt.gca()
        ax.set_aspect("equal")
        ax.set(
            xlabel=f"x ({trj.__dict__.get('spatial_units','m')})",
            ylabel=f"y ({trj.__dict__.get('spatial_units','m')})",
            title="Rolling Trajectory Convex Hull\nWindow={window},Step={step}",
        )


def plot_period(trj: traja.TrajaDataFrame, col="x", dark=(7, 19), **kwargs):
    time_col = traja._get_time_col(trj)
    _trj = trj.set_index(time_col)
    if not col in _trj:
        raise ValueError(f"{col} not a column in dataframe")
    series = _trj[col]
    fig, ax = plt.subplots()
    series.plot(ax=ax)

    dates = np.unique(series.index.date)

    nights = []
    nights.append([(date, date + timedelta(hours=dark[0])) for date in dates])
    nights.append(
        [(date + timedelta(hours=dark[1]), date + timedelta(days=1)) for date in dates]
    )
    for interval in nights:
        t0, t1 = interval
        ax.axvspan(t0, t1, color="gray", alpha=0.2)

    # Format date displayed on the x axis
    xfmt = md.DateFormatter("%H:%M\n%m-%d-%y")
    ax.xaxis.set_major_formatter(xfmt)

    if kwargs.get("interactive"):
        plt.show()


def plot_rolling_hull_3d(trj: TrajaDataFrame, window=100, step=20, **kwargs):
    from mpl_toolkits.mplot3d import Axes3D

    hulls = []

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")

    for offset, wind in _rolling(trj, window=window, step=step):
        if wind.dropna().empty:
            continue
        shape = wind.traja.to_shapely()
        hull = shape.convex_hull
        hulls.append(hull)

    xlim, ylim = traja.trajectory._get_xylim(trj)
    plt.xlim = xlim
    plt.ylim = ylim
    outlines = []
    for idx, hull in enumerate(hulls):
        if hasattr(hull, "exterior"):  # Occassionally a Point object without it reaches
            outlines.append(np.array(hull.exterior.xy))

    # Add plots to axes
    NLINES = len(outlines)
    cm = plt.get_cmap(kwargs.get("cmap", "plasma"))
    ax.set_prop_cycle(color=[cm(1.0 * i / (NLINES)) for i in range(NLINES)])
    for z, xy in enumerate(outlines):
        ax.plot(*xy, z)

    ax.set(
        xlabel=f"{trj.__dict__.get('spatial_units','m')}",
        ylabel=f"{trj.__dict__.get('spatial_units','m')}",
        title=f"Rolling Trajectory Convex Hull\nWindow={window},Step={step}",
    )

    if kwargs.get("interactive"):
        plt.show()


def plot_3d(trj: TrajaDataFrame, **kwargs) -> matplotlib.collections.PathCollection:
    """Plot 3D trajectory for single identity over period.

    Args:
      trj (:class:`traja.TrajaDataFrame`): trajectory
      n_coords (int, optional): Number of coordinates to plot
      **kwargs: additional keyword arguments to :meth:`matplotlib.axes.Axes.scatter`

    Returns:
        ax (:class:`~matplotlib.collections.PathCollection`): Axes of plot

    .. note::
        Takes a while to plot large trajectories. Consider using first::
        
            rt = trj.traja.rediscretize(R=1.) # Replace R with appropriate step length
            rt.traja.plot_3d()

    """
    from mpl_toolkits.mplot3d import Axes3D

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.set_xlabel("x", fontsize=15)
    ax.set_zlabel("time", fontsize=15)
    ax.set_ylabel("y", fontsize=15)
    title = kwargs.pop("title", "Trajectory")
    ax.set_title(f"{title}", fontsize=20)
    ax.plot(trj.x, trj.y, trj.index)
    cmap = kwargs.pop("cmap", "winter")
    cm = plt.get_cmap(cmap)
    NPOINTS = len(trj)
    ax.set_prop_cycle(color=[cm(1.0 * i / (NPOINTS - 1)) for i in range(NPOINTS - 1)])
    for i in range(NPOINTS - 1):
        ax.plot(trj.x[i : i + 2], trj.y[i : i + 2], trj.index[i : i + 2])

    dist = kwargs.pop("dist", None)
    if dist:
        ax.dist = dist
    labelpad = kwargs.pop("labelpad", None)
    if labelpad:
        from matplotlib import rcParams

        rcParams["axes.labelpad"] = labelpad

    return ax


def plot(
    trj: TrajaDataFrame,
    n_coords: Optional[int] = None,
    show_time: bool = False,
    accessor: Optional[traja.TrajaAccessor] = None,
    ax=None,
    **kwargs,
) -> matplotlib.collections.PathCollection:
    """Plot trajectory for single animal over period.

    Args:
      trj (:class:`traja.TrajaDataFrame`): trajectory
      n_coords (int, optional): Number of coordinates to plot
      show_time (bool): Show colormap as time
      accessor (:class:`~traja.accessor.TrajaAccessor`, optional): TrajaAccessor instance
      ax (:class:`~matplotlib.axes.Axes`): axes for plotting
      interactive (bool): show plot immediately
      **kwargs: additional keyword arguments to :meth:`matplotlib.axes.Axes.scatter`

    Returns:
        collection (:class:`~matplotlib.collections.PathCollection`): collection that was plotted

    """
    import matplotlib.patches as patches
    from matplotlib.path import Path

    after_plot_args, kwargs = _get_after_plot_args(**kwargs)

    GRAY = "#999999"

    xlim = kwargs.pop("xlim", None)
    ylim = kwargs.pop("ylim", None)
    if not xlim or not ylim:
        xlim, ylim = traja.trajectory._get_xylim(trj)

    title = kwargs.pop("title", None)
    time_units = kwargs.pop("time_units", "s")
    fps = kwargs.pop("fps", None)
    figsize = kwargs.pop("figsize", None)

    coords = trj[["x", "y"]]
    time_col = traja.trajectory._get_time_col(trj)

    if time_col == "index":
        is_datetime = True
    else:
        is_datetime = is_datetime64_any_dtype(trj[time_col]) if time_col else False

    if n_coords is None:
        # Plot all coords
        start, end = 0, len(coords)
        verts = coords.iloc[start:end].values
    else:
        # Plot first `n_coords`
        verts = coords.iloc[:n_coords].values

    n_coords = len(verts)

    codes = [Path.MOVETO] + [Path.LINETO] * (len(verts) - 1)
    path = Path(verts, codes)

    if not ax:
        fig, ax = plt.subplots(figsize=figsize)
        fig.canvas.draw()

    patch = patches.PathPatch(path, edgecolor=GRAY, facecolor="none", lw=3, alpha=0.3)
    ax.add_patch(patch)

    xs, ys = zip(*verts)

    if time_col == "index":
        # DatetimeIndex determines color
        colors = [ind for ind, x in enumerate(trj.index[:n_coords])]
    elif time_col and time_col != "index":
        # `time_col` determines color
        colors = [ind for ind, x in enumerate(trj[time_col].iloc[:n_coords])]
    else:
        # Frame count determines color
        colors = trj.index[:n_coords]

    if time_col:
        # TODO: Calculate fps if not in datetime
        vmin = min(colors)
        vmax = max(colors)
        if is_datetime:
            # Show timestamps without units
            time_units = ""
    else:
        # Index/frame count is our only reference
        vmin = trj.index[0]
        vmax = trj.index[n_coords - 1]
        if not show_time:
            time_units = ""
    label = f"Time ({time_units})" if time_units else ""

    collection = ax.scatter(
        xs,
        ys,
        c=colors,
        s=kwargs.pop("s", 5),
        cmap=plt.cm.viridis,
        alpha=0.7,
        vmin=vmin,
        vmax=vmax,
        **kwargs,
    )

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    if kwargs.pop("invert_yaxis", None):
        plt.gca().invert_yaxis()

    _label_axes(trj, ax)
    ax.set_title(title)
    ax.set_aspect("equal")

    # Number of color bar ticks
    CBAR_TICKS = 10 if n_coords > 20 else n_coords
    indices = np.linspace(0, n_coords - 1, CBAR_TICKS, endpoint=True, dtype=int)
    cbar = plt.colorbar(
        collection, fraction=0.046, pad=0.04, orientation="vertical", label=label
    )

    # Get colorbar labels from time
    if time_col == "index":
        if is_datetime64_any_dtype(trj.index):
            cbar_labels = (
                trj.index[indices].strftime("%Y-%m-%d %H:%M:%S").values.astype(str)
            )
        elif is_timedelta64_dtype(trj.index):
            if time_units in ("s", "", None):
                cbar_labels = [round(x, 2) for x in trj.index[indices].total_seconds()]
            else:
                print("Time unit {} not yet implemented".format(time_units))
        else:
            raise NotImplementedError(
                "Indexing on {} is not yet implemented".format(type(trj.index))
            )
    elif time_col and is_timedelta64_dtype(trj[time_col]):
        cbar_labels = trj[time_col].iloc[indices].dt.total_seconds().values
        cbar_labels = ["%.2f" % number for number in cbar_labels]
    elif time_col and is_datetime:
        cbar_labels = (
            trj[time_col]
            .iloc[indices]
            .dt.strftime("%Y-%m-%d %H:%M:%S")
            .values.astype(str)
        )
    else:
        # Convert frames to time
        if time_col:
            cbar_labels = trj[time_col][indices].values
        else:
            cbar_labels = trj.index[indices].values
        cbar_labels = np.round(cbar_labels, 6)
        if fps is not None and fps > 0 and fps is not 1 and show_time:
            cbar_labels = cbar_labels / fps

    cbar.set_ticks(indices)
    cbar.set_ticklabels(cbar_labels)
    plt.tight_layout()

    _process_after_plot_args(**after_plot_args)
    return collection


def plot_collection(
    trjs: Union[pd.DataFrame, traja.TrajaDataFrame],
    id_col: str = "id",
    colors: Optional[Union[dict, List[str]]] = None,
    **kwargs,
):
    """Plot trajectories of multiple subjects identified by `id`.

    Args:
        trjs: dataframe with multiple trajectories
        id_col: name of id_col, default is "id"
        colors (Optional): color lookup matching substrings to discreet colors. Possible values are, eg:
                                - {"car0":"red","car1":"blue"}
                                - {"car":"red","person":blue"}
                                - ["car", "person"]
        kwargs: kwargs to :meth:`matplotlib.axes.Axes.plot`

    Returns:
        lines (list of `~matplotlib.lines.Line2D` objects): lines of plot

    """
    ids = trjs[id_col].unique()

    # Get plot keyword args
    colormap = kwargs.pop("cmap", "hsv")
    alpha = kwargs.pop("alpha", 0.2)
    linestyle = kwargs.pop("linestyle", "-")
    marker = kwargs.pop("marker", "o")

    labels = [None] * len(ids)

    if not colors:
        cmap = plt.cm.get_cmap(colormap, lut=len(ids) if len(ids) > 1 else None)
        colors = [cmap(idx) for idx in range(len(ids))]
    elif isinstance(colors, list):
        cmap = plt.cm.get_cmap(colormap, len(colors))
        color_lookup = []
        for ind, id in enumerate(ids):
            for idx, substring in enumerate(colors):
                if substring in id:
                    color_lookup.append(cmap(idx))
                    labels[ind] = substring
                    break
            else:
                raise Exception(f"No substring matching {id} in {colors}.")
        colors = color_lookup
    elif isinstance(colors, dict):
        color_lookup = [colors.get(id) for id in ids]
        colors = color_lookup
        labels = ids

    fig, ax = plt.subplots()
    lines = []
    for idx, id in enumerate(ids):
        trj = trjs[trjs[id_col] == id]
        l = ax.plot(
            trj.x,
            trj.y,
            linestyle=linestyle,
            marker=marker,
            c=colors[idx],
            alpha=alpha,
            label=labels[idx],
            **kwargs,
        )
        lines.extend(l)

    handles, labels = plt.gca().get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    plt.legend(
        by_label.values(),
        by_label.keys(),
        bbox_to_anchor=(1.05, 1),
        loc=2,
        borderaxespad=0.0,
    )
    plt.tight_layout()
    return lines


def _label_axes(trj: TrajaDataFrame, ax) -> Axes:
    if "spatial_units" in trj.__dict__:
        ax.set_xlabel(trj.spatial_units)
        ax.set_ylabel(trj.spatial_units)
    return ax


def plot_quiver(
    trj: TrajaDataFrame,
    bins: Optional[Union[int, tuple]] = None,
    quiverplot_kws: dict = {},
    **kwargs,
) -> Axes:
    """Plot average flow from each grid cell to neighbor.

    Args:
        bins (int or tuple): Tuple of x,y bin counts; if `bins` is int, bin count of x,
                                with y inferred from aspect ratio
        quiverplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.quiver`

    Returns:
        ax (:class:`~matplotlib.axes.Axes`): Axes of quiver plot
    """

    after_plot_args, _ = _get_after_plot_args(**kwargs)

    X, Y, U, V = coords_to_flow(trj, bins)
    Z = np.sqrt(U * U + V * V)

    fig, ax = plt.subplots()

    ax.quiver(X, Y, U, V, units="width", **quiverplot_kws)
    ax = _label_axes(trj, ax)
    ax.set_aspect("equal")

    _process_after_plot_args(**after_plot_args)
    return ax


def plot_contour(
    trj: TrajaDataFrame,
    bins: Optional[Union[int, tuple]] = None,
    filled: bool = True,
    quiver: bool = True,
    contourplot_kws: dict = {},
    contourfplot_kws: dict = {},
    quiverplot_kws: dict = {},
    **kwargs,
) -> Axes:
    """Plot average flow from each grid cell to neighbor.

    Args:
        bins (int or tuple): Tuple of x,y bin counts; if `bins` is int, bin count of x,
                                with y inferred from aspect ratio
        contourplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.contour`
        contourfplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.contourf`
        quiverplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.quiver`

    Returns:
        ax (:class:`~matplotlib.axes.Axes`): Axes of quiver plot
    """

    after_plot_args, _ = _get_after_plot_args(**kwargs)

    X, Y, U, V = coords_to_flow(trj, bins)
    Z = np.sqrt(U * U + V * V)

    fig, ax = plt.subplots()

    if filled:
        cfp = plt.contourf(X, Y, Z, **contourfplot_kws)
        plt.colorbar(cfp, ax=ax)
    plt.contour(
        X, Y, Z, colors="k", linewidths=1, linestyles="solid", **contourplot_kws
    )
    if quiver:
        ax.quiver(X, Y, U, V, units="width", **quiverplot_kws)

    ax = _label_axes(trj, ax)
    ax.set_aspect("equal")

    _process_after_plot_args(**after_plot_args)
    return ax


def plot_surface(
    trj: TrajaDataFrame,
    bins: Optional[Union[int, tuple]] = None,
    cmap: str = "jet",
    **surfaceplot_kws: dict,
) -> Figure:
    """Plot surface of flow from each grid cell to neighbor in 3D.

    Args:
        bins (int or tuple): Tuple of x,y bin counts; if `bins` is int, bin count of x,
                                with y inferred from aspect ratio
        cmap (str): color map
        surfaceplot_kws: Additional keyword arguments for :meth:`~mpl_toolkits.mplot3D.Axes3D.plot_surface`

    Returns:
        ax (:class:`~matplotlib.axes.Axes`): Axes of quiver plot
    """
    from mpl_toolkits.mplot3d import Axes3D

    after_plot_args, surfaceplot_kws = _get_after_plot_args(**surfaceplot_kws)

    X, Y, U, V = coords_to_flow(trj, bins)
    Z = np.sqrt(U * U + V * V)

    fig = plt.figure()
    ax = fig.gca(projection="3d")
    ax.plot_surface(
        X, Y, Z, cmap=matplotlib.cm.coolwarm, linewidth=0, **surfaceplot_kws
    )

    ax = _label_axes(trj, ax)
    try:
        ax.set_aspect("equal")
    except NotImplementedError:
        # 3D
        pass

    _process_after_plot_args(**after_plot_args)
    return ax


def plot_stream(
    trj: TrajaDataFrame,
    bins: Optional[Union[int, tuple]] = None,
    cmap: str = "jet",
    contourfplot_kws: dict = {},
    contourplot_kws: dict = {},
    streamplot_kws: dict = {},
    **kwargs,
) -> Figure:
    """Plot average flow from each grid cell to neighbor.

    Args:
        bins (int or tuple): Tuple of x,y bin counts; if `bins` is int, bin count of x,
                                with y inferred from aspect ratio
        contourplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.contour`
        contourfplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.contourf`
        streamplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.streamplot`

    Returns:
        ax (:class:`~matplotlib.axes.Axes`): Axes of stream plot

    """

    after_plot_args, _ = _get_after_plot_args(**kwargs)
    X, Y, U, V = coords_to_flow(trj, bins)
    Z = np.sqrt(U * U + V * V)

    fig, ax = plt.subplots()

    plt.contourf(X, Y, Z, **contourfplot_kws)
    plt.contour(
        X, Y, Z, colors="k", linewidths=1, linestyles="solid", **contourplot_kws
    )
    ax.streamplot(X, Y, U, V, color=Z, cmap=cmap, **streamplot_kws)

    ax = _label_axes(trj, ax)
    ax.set_aspect("equal")

    _process_after_plot_args(**after_plot_args)
    return ax


def plot_flow(
    trj: TrajaDataFrame,
    kind: str = "quiver",
    *args,
    contourplot_kws: dict = {},
    contourfplot_kws: dict = {},
    streamplot_kws: dict = {},
    quiverplot_kws: dict = {},
    surfaceplot_kws: dict = {},
    **kwargs,
) -> Figure:
    """Plot average flow from each grid cell to neighbor.

    Args:
        bins (int or tuple): Tuple of x,y bin counts; if `bins` is int, bin count of x,
                                with y inferred from aspect ratio
        kind (str): Choice of 'quiver','contourf','stream','surface'. Default is 'quiver'.
        contourplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.contour`
        contourfplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.contourf`
        streamplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.streamplot`
        quiverplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.quiver`
        surfaceplot_kws: Additional keyword arguments for :meth:`~matplotlib.axes.Axes.plot_surface`

    Returns:
        ax (:class:`~matplotlib.axes.Axes`): Axes of plot
    """
    if kind == "quiver":
        return plot_quiver(trj, *args, **quiverplot_kws, **kwargs)
    elif kind == "contour":
        return plot_contour(trj, filled=False, *args, **quiverplot_kws, **kwargs)
    elif kind == "contourf":
        return plot_contour(trj, *args, **quiverplot_kws, **kwargs)
    elif kind == "stream":
        return plot_stream(
            trj,
            *args,
            contourplot_kws=contourplot_kws,
            contourfplot_kws=contourfplot_kws,
            streamplot_kws=streamplot_kws,
            **kwargs,
        )
    elif kind == "surface":
        return plot_surface(trj, *args, **surfaceplot_kws, **kwargs)
    else:
        raise NotImplementedError(f"Kind {kind} is not implemented.")


def _get_after_plot_args(**kwargs: dict) -> (dict, dict):
    after_plot_args = dict(
        interactive=kwargs.pop("interactive", True),
        filepath=kwargs.pop("filepath", None),
    )
    return after_plot_args, kwargs


def trip_grid(
    trj: TrajaDataFrame,
    bins: Union[tuple, int] = 10,
    log: bool = False,
    spatial_units: str = None,
    normalize: bool = False,
    hist_only: bool = False,
    **kwargs,
) -> Tuple[np.ndarray, PathCollection]:
    """Generate a heatmap of time spent by point-to-cell gridding.

    Args:
      bins (int, optional): Number of bins (Default value = 10)
      log (bool): log scale histogram (Default value = False)
      spatial_units (str): units for plotting
      normalize (bool): normalize histogram into density plot
      hist_only (bool): return histogram without plotting

    Returns:
        hist (:class:`numpy.ndarray`): 2D histogram as array
        image (:class:`matplotlib.collections.PathCollection`: image of histogram

    """
    after_plot_args, kwargs = _get_after_plot_args(**kwargs)

    bins = traja.trajectory._bins_to_tuple(trj, bins)
    # TODO: Add kde-based method for line-to-cell gridding
    df = trj[["x", "y"]].dropna()

    # Set aspect if `xlim` and `ylim` set.
    if "xlim" in kwargs and "ylim" in kwargs:
        xlim, ylim = kwargs.pop("xlim"), kwargs.pop("ylim")
    else:
        xlim, ylim = traja.trajectory._get_xylim(df)
    xmin, xmax = xlim
    ymin, ymax = ylim

    x, y = zip(*df.values)
    # FIXME: Remove redundant histogram calculation
    hist, x_edges, y_edges = np.histogram2d(
        x, y, bins, range=((xmin, xmax), (ymin, ymax)), normed=normalize
    )

    if log:
        hist = np.log(hist + np.e)
    if hist_only:  # TODO: Evaluate potential use cases or remove
        return (hist, None)
    fig, ax = plt.subplots()

    image = ax.imshow(
        hist, interpolation="bilinear", aspect="equal", extent=[xmin, xmax, ymin, ymax]
    )
    # TODO: Adjust colorbar ytick_labels to correspond with time
    label = "Frames" if not log else "$ln(frames)$"
    plt.colorbar(image, ax=ax, label=label)

    _label_axes(trj, ax)

    plt.title("Time spent{}".format(" (Logarithmic)" if log else ""))

    _process_after_plot_args(**after_plot_args)
    # TODO: Add method for most common locations in grid
    # peak_index = unravel_index(hist.argmax(), hist.shape)
    return hist, image


def _process_after_plot_args(**after_plot_args):
    filepath = after_plot_args.get("filepath")
    if filepath:
        plt.savefig(filepath)


def color_dark(
    series: pd.Series, ax: matplotlib.axes.Axes, start: int = 19, end: int = 7
):
    """Color dark phase in plot."""
    assert is_datetime_or_timedelta_dtype(
        series.index
    ), f"Series must have datetime index but has {type(series.index)}"

    dark_mask = (series.index.hour >= start) | (series.index.hour < end)
    run_values, run_starts, run_lengths = find_runs(dark_mask)

    for idx, is_dark in enumerate(run_values):
        if is_dark:
            start = run_starts[idx]
            end = run_starts[idx] + run_lengths[idx] - 1
            ax.axvspan(series.index[start], series.index[end], alpha=0.5, color="gray")


def find_runs(x: pd.Series) -> (np.ndarray, np.ndarray, np.ndarray):
    """Find runs of consecutive items in an array.
    From https://gist.github.com/alimanfoo/c5977e87111abe8127453b21204c1065."""

    # ensure array
    x = np.asanyarray(x)
    if x.ndim != 1:
        raise ValueError("only 1D array supported")
    n = x.shape[0]

    # handle empty array
    if n == 0:
        return np.array([]), np.array([]), np.array([])

    else:
        # find run starts
        loc_run_start = np.empty(n, dtype=bool)
        loc_run_start[0] = True
        np.not_equal(x[:-1], x[1:], out=loc_run_start[1:])
        run_starts = np.nonzero(loc_run_start)[0]

        # find run values
        run_values = x[loc_run_start]

        # find run lengths
        run_lengths = np.diff(np.append(run_starts, n))

        return run_values, run_starts, run_lengths


def fill_ci(series: pd.Series, window: Union[int, str]) -> Figure:
    """Fill confidence interval defined by SEM over mean of `window`. Window can be interval or offset, eg, '30s'."""
    assert is_datetime_or_timedelta_dtype(
        series.index
    ), f"Series index must be datetime but is {type(series.index)}"
    smooth_path = series.rolling(window).mean()
    path_deviation = series.rolling(window).std()

    fig, ax = plt.subplots()

    plt.plot(smooth_path.index, smooth_path, "b")
    plt.fill_between(
        path_deviation.index,
        (smooth_path - 2 * path_deviation),
        (smooth_path + 2 * path_deviation),
        color="b",
        alpha=0.2,
    )

    plt.gcf().autofmt_xdate()
    return ax


def plot_xy(xy: np.ndarray, *args: Optional, **kwargs: Optional):
    """Plot trajectory from xy values.

    Args:

        xy (np.ndarray) : xy values of dimensions N x 2
        *args           : Plot args
        **kwargs        : Plot kwargs
    """
    trj = traja.from_xy(xy)
    trj.traja.plot(*args, **kwargs)


def plot_actogram(
    series: pd.Series, dark=(19, 7), ax: matplotlib.axes.Axes = None, **kwargs
):
    """Plot activity or displacement as an actogram.

    .. note::

       For published example see Eckel-Mahan K, Sassone-Corsi P. Phenotyping Circadian Rhythms in Mice.
       Curr Protoc Mouse Biol. 2015;5(3):271-281. Published 2015 Sep 1. doi:10.1002/9780470942390.mo140229

    """
    after_plot_args, _ = _get_after_plot_args(**kwargs)
    assert is_datetime_or_timedelta_dtype(
        series.index
    ), f"Series must have datetime index but has {type(series.index)}"

    ax = series.plot(ax=ax)
    ax.set_ylabel(series.name)

    color_dark(series, ax, start=dark[0], end=dark[1])

    _process_after_plot_args(**after_plot_args)


def _polar_bar(
    radii: np.ndarray,
    theta: np.ndarray,
    bin_size: int = 2,
    ax: Optional[matplotlib.axes.Axes] = None,
    overlap: bool = True,
    **kwargs: str,
) -> Axes:
    after_plot_args, kwargs = _get_after_plot_args(**kwargs)

    title = kwargs.pop("title", None)
    ax = ax or plt.subplot(111, projection="polar")

    hist, bin_edges = np.histogram(
        theta, bins=np.arange(-180, 180 + bin_size, bin_size)
    )
    centers = np.deg2rad(np.ediff1d(bin_edges) // 2 + bin_edges[:-1])

    radians = np.deg2rad(theta)

    width = np.deg2rad(bin_size)
    angle = radians if overlap else centers
    height = radii if overlap else hist
    max_height = max(height)
    bars = ax.bar(angle, height, width=width, bottom=0.0, **kwargs)
    for h, bar in zip(height, bars):
        bar.set_facecolor(plt.cm.jet(h / max_height))
        bar.set_alpha(0.5)
    if isinstance(ax, matplotlib.axes.Axes):
        ax.set_theta_zero_location("N")
        ax.set_xticklabels(["0", "45", "90", "135", "180", "-135", "-90", "-45"])
    if title:
        plt.title(title + "\n", y=1.08)
    plt.tight_layout()

    _process_after_plot_args(**after_plot_args)
    return ax


def polar_bar(
    trj: TrajaDataFrame,
    feature: str = "turn_angle",
    bin_size: int = 2,
    overlap: bool = True,
    ax: Optional[matplotlib.axes.Axes] = None,
    **plot_kws: str,
) -> Axes:
    """Plot polar bar chart.

    Args:
        trj (:class:`traja.TrajaDataFrame`): trajectory
        feature (str): Options: 'turn_angle', 'heading'
        bins (int): width of bins
        overlap (bool): Overlapping shows all values, if set to false is a histogram

    Returns:
        ax (:class:`~matplotlib.collections.PathCollection`): Axes of plot

    """
    DIST_THRESHOLD = 0.001
    # Get displacement

    displacement = traja.trajectory.calc_displacement(trj)
    trj["displacement"] = displacement
    trj = trj.loc[trj.displacement > DIST_THRESHOLD]
    if feature == "turn_angle":
        feature_series = traja.trajectory.calc_turn_angle(trj)
        trj["turn_angle"] = feature_series
        trj.turn_angle = trj.turn_angle.shift(-1)
    elif feature == "heading":
        feature_series = traja.trajectory.calc_heading(trj)
        trj[feature] = feature_series

    trj = trj[pd.notnull(trj[feature])]
    trj = trj[pd.notnull(trj.displacement)]

    assert len(trj) > 0, "Dataframe is empty after filtering, check coordinates"

    ax = _polar_bar(
        trj.displacement,
        trj[feature],
        bin_size=bin_size,
        overlap=overlap,
        ax=ax,
        **plot_kws,
    )
    return ax


def plot_clustermap(
    displacements: List[pd.Series],
    rule: Optional[str] = None,
    nr_steps=None,
    colors: Optional[List[Union[int, str]]] = None,
    **kwargs,
):
    """Plot cluster map / dendrogram of trajectories with DatetimeIndex.

    Args:
        displacements: list of pd.Series, outputs of :func:`traja.calc_displacement()`
        rule:   how to resample series, eg '30s' for 30-seconds
        nr_steps: select first N samples for clustering
        colors: list of colors (eg, 'b','r') to map to each trajectory
        kwargs: keyword arguments for :func:`seaborn.clustermap`

    Returns:
        cg: a :func:`seaborn.matrix.ClusterGrid` instance

    .. note::

        Requires seaborn to be installed. Install it with 'pip install seaborn'.

    """
    try:
        import seaborn as sns
    except ImportError:
        print("seaborn is not installed. Install it with 'pip install seaborn'")
        return

    after_plot_args, _ = _get_after_plot_args(**kwargs)

    series_lst = []
    for disp in displacements:
        if rule:
            disp = disp.resample(rule).sum()
        series_lst.append(disp)

    df = pd.DataFrame(series_lst)
    df.columns = range(len(df.columns))
    df.reset_index(drop=True, inplace=True)

    if not nr_steps:
        nr_steps = df.shape[1]

    cg = sns.clustermap(
        df.fillna(0).iloc[:, :nr_steps],
        xticklabels=False,
        col_cluster=False,
        figsize=(16, 6),
        cmap="Greys",
        row_colors=colors,
        **kwargs,
    )
    plt.setp(cg.ax_heatmap.yaxis.get_majorticklabels(), rotation=0)

    _process_after_plot_args(**after_plot_args)
    return cg


def _get_markov_edges(Q: pd.DataFrame, greater_than=0.1):
    """Select edges greater than a threshold of weight."""
    edges = {}
    for col in Q.columns:
        for idx in Q.index:
            if greater_than and Q.loc[idx, col] > greater_than:
                edges[(idx, col)] = Q.loc[idx, col]
    return edges


def plot_transition_graph(
    data: Union[pd.DataFrame, traja.TrajaDataFrame, np.ndarray],
    outpath="markov.dot",
    interactive=True,
):
    """Plot transition graph with networkx.

    Args:
        data (trajectory or transition_matrix)
    
    .. note::        
        Modified from http://www.blackarbs.com/blog/introduction-hidden-markov-models-python-networkx-sklearn/2/9/2017
    
    """
    try:
        import networkx as nx
        import pydot
        import graphviz
    except ImportError as e:
        raise ImportError(f"{e} - please install it with pip")

    if (
        isinstance(data, (traja.TrajaDataFrame))
        or isinstance(data, pd.DataFrame)
        and "x" in data
    ):
        transition_matrix = traja.transitions(data)
        edges_wts = _get_markov_edges(pd.DataFrame(transition_matrix))
        states_ = list(range(transition_matrix.shape[0]))

    # create graph object
    G = nx.MultiDiGraph()

    # nodes correspond to states
    G.add_nodes_from(states_)

    # edges represent transition probabilities
    for k, v in edges_wts.items():
        tmp_origin, tmp_destination = k[0], k[1]
        G.add_edge(tmp_origin, tmp_destination, weight=v.round(4), label=v.round(4))

    pos = nx.drawing.nx_pydot.graphviz_layout(G, prog="dot")
    nx.draw_networkx(G, pos)

    # create edge labels for jupyter plot but is not necessary
    edge_labels = {(n1, n2): d["label"] for n1, n2, d in G.edges(data=True)}
    nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)
    if os.exists(outpath):
        print(f"Overwriting {outpath}")
    nx.drawing.nx_pydot.write_dot(G, outpath)

    if interactive:
        # Plot
        from graphviz import Source

        s = Source.from_file(outpath)
        s.view()


def plot_transition_matrix(
    data: Union[pd.DataFrame, traja.TrajaDataFrame, np.ndarray],
    interactive=True,
    **kwargs,
) -> matplotlib.image.AxesImage:
    """Plot transition matrix.
    
    Args:
        data (trajectory or square transition matrix)
        interactive (bool): show plot
        kwargs: kwargs to :func:`traja.grid_coordinates`

    Returns:
        axesimage (matplotlib.image.AxesImage)
    
    """
    if isinstance(data, np.ndarray):
        if data.shape[0] != data.shape[1]:
            raise ValueException(
                f"Ndarray input must be square transition matrix, shape is {data.shape}"
            )
        transition_matrix = data
    elif isinstance(data, (pd.DataFrame, traja.TrajaDataFrame)):
        transition_matrix = traja.transitions(data, **kwargs)
    img = plt.imshow(transition_matrix)
    if interactive:
        plt.show()
    return img


def animate(trj: TrajaDataFrame, polar: bool = True, save: bool = False):
    """Animate trajectory.

    Args:
        polar (bool): include polar bar chart with turn angle
        save (bool): save video to ``trajectory.mp4``

    """
    from matplotlib import animation
    from matplotlib.animation import FuncAnimation

    displacement = traja.trajectory.calc_displacement(trj).reset_index(drop=True)
    # heading = traja.calc_heading(trj)
    turn_angle = traja.trajectory.calc_turn_angle(trj).reset_index(drop=True)
    xy = trj[["x", "y"]].reset_index(drop=True)

    POLAR_STEPS = XY_STEPS = 20
    DISPLACEMENT_THRESH = 0.025
    bin_size = 2
    overlap = True

    fig = plt.figure(figsize=(8, 6))
    ax1 = plt.subplot(211)

    fig.add_subplot(ax1)
    if polar:
        ax2 = plt.subplot(212, polar="projection")
        ax2.set_theta_zero_location("N")
        ax2.set_xticklabels(["0", "45", "90", "135", "180", "-135", "-90", "-45"])
        fig.add_subplot(ax2)
        ax2.bar(
            np.zeros(XY_STEPS), np.zeros(XY_STEPS), width=np.zeros(XY_STEPS), bottom=0.0
        )

    xlim, ylim = traja.trajectory._get_xylim(trj)
    ax1.set(
        xlim=xlim,
        ylim=ylim,
        ylabel=trj.spatial_units or "m",
        xlabel=trj.spatial_units or "m",
        aspect="equal",
    )

    alphas = np.linspace(0.1, 1, XY_STEPS)
    rgba_colors = np.zeros((XY_STEPS, 4))
    rgba_colors[:, 0] = 1.0  # red
    rgba_colors[:, 3] = alphas
    scat = ax1.scatter(
        range(XY_STEPS), range(XY_STEPS), marker=".", color=rgba_colors[:XY_STEPS]
    )

    def update(frame_number):
        ind = frame_number % len(xy)
        if ind < XY_STEPS:
            scat.set_offsets(xy[:XY_STEPS])
        else:
            prev_steps = max(ind - XY_STEPS, 0)
            scat.set_offsets(xy[prev_steps:ind])

        displacement_str = (
            rf"$\bf{displacement[ind]:.2f}$"
            if displacement[ind] >= DISPLACEMENT_THRESH
            else f"{displacement[ind]:.2f}"
        )

        x, y = xy.iloc[ind]
        ax1.set_title(
            f"frame {ind} - distance (cm/0.25s): {displacement_str}\n"
            f"x: {x:.2f}, y: {y:.2f}\n"
            f"turn_angle: {turn_angle[ind]:.2f}"
        )

        if polar and ind > 1:
            ax2.clear()
            start_index = max(ind - POLAR_STEPS, 0)

            theta = turn_angle[start_index:ind]
            radii = displacement[start_index:ind]

            hist, bin_edges = np.histogram(
                theta, bins=np.arange(-180, 180 + bin_size, bin_size)
            )
            centers = np.deg2rad(np.ediff1d(bin_edges) // 2 + bin_edges[:-1])

            radians = np.deg2rad(theta)

            width = np.deg2rad(bin_size)
            angle = radians if overlap else centers
            height = radii if overlap else hist
            max_height = displacement.max() if overlap else max(hist)

            bars = ax2.bar(angle, height, width=width, bottom=0.0)
            for idx, (h, bar) in enumerate(zip(height, bars)):
                bar.set_facecolor(plt.cm.jet(h / max_height))
                bar.set_alpha(0.8 * (idx / POLAR_STEPS))
            ax2.set_theta_zero_location("N")
            ax2.set_xticklabels(["0", "45", "90", "135", "180", "-135", "-90", "-45"])
        plt.tight_layout()

    anim = FuncAnimation(fig, update, interval=10, frames=range(len(xy)))
    if save:
        try:
            anim.save("trajectory.mp4", writer=animation.FFMpegWriter(fps=10))
        except FileNotFoundError:
            raise Exception("FFmpeg not installed, please install it.")
    else:
        plt.show()



import os

import numpy as np
import pandas as pd

import traja


df = traja.generate(n=20)


def test_from_df():
    df = pd.DataFrame({"x": [1, 2, 3], "y": [2, 3, 4]})
    trj = traja.parsers.from_df(df)
    np.testing.assert_allclose(df, trj)
    assert isinstance(trj, traja.TrajaDataFrame)


def test_read_file():
    datapath = os.path.join(traja.__path__[0], "tests", "data", "3527.csv")
    trj = traja.parsers.read_file(datapath)
    assert isinstance(trj, traja.TrajaDataFrame)
    assert "Frame" in trj
    assert "Time" in trj
    assert "TrackId" in trj
    assert "x" in trj
    assert "y" in trj
    assert "ValueChanged" in trj
